@@ -670,16 +670,18 @@ PROCESS is the process under watch, EVENT is the event occurred."
       (<= line 2))))
 
 (defun tabnine--response-display-with-capf-p(response)
-  "Test RESPONSE whether display with capf."
+  "Test RESPONSE whether display with capf, prefer display with overlay."
   (when response
     (let* ((display-with-overlay)
 	   (results (plist-get response :results)))
-      (mapc (lambda(x)
-	      (unless (tabnine--completion-capf-p x)
-		(setq display-with-overlay t))) results)
-      (when (not tabnine--trigger-with-capf)
+      (if tabnine--trigger-with-capf
+	  (catch 'break
+	    (mapc (lambda(x)
+		    (unless (tabnine--completion-capf-p x)
+		      (setq display-with-overlay t)
+		      (throw 'break nil))) results))
 	(setq display-with-overlay t))
-      (or (not tabnine-mode) (not display-with-overlay)))))
+      (not (and tabnine-mode display-with-overlay)))))
 
 (defun tabnine--response-latest-completion-p(response)
   "Test RESPONSE if the latest completion."
@@ -689,16 +691,16 @@ PROCESS is the process under watch, EVENT is the event occurred."
 (defun tabnine--filter-completions(completions)
   "Filter duplicates and bad COMPLETIONS result."
   (when completions
-    (when-let ((completions (cl-remove-duplicates completions
-						  :key (lambda (x) (plist-get x :new_prefix))
-						  :test #'s-equals-p))
-	       (completions (cl-remove-duplicates completions
-						  :key (lambda (x) (let ((new_prefix (plist-get x :new_prefix))
-									 (new_posfix (plist-get x :new_posfix)))
-								     (s-trim (concat new_prefix (or new_posfix "")))))
-						  :test #'s-equals-p))
-	       (completions (cl-remove-if #'tabnine--invalid-completion-p
-					  completions)))
+    (when-let* ((completions (cl-remove-duplicates completions
+						   :key (lambda (x) (plist-get x :new_prefix))
+						   :test #'s-equals-p))
+		(completions (cl-remove-duplicates completions
+						   :key (lambda (x) (let ((new_prefix (plist-get x :new_prefix))
+								     (new_posfix (plist-get x :new_posfix)))
+								 (s-trim (concat new_prefix (or new_posfix "")))))
+						   :test #'s-equals-p))
+		(completions (cl-remove-if #'tabnine--invalid-completion-p
+					   completions)))
       completions)))
 
 (defun tabnine--process-filter (_ output)
@@ -721,11 +723,24 @@ PROCESS is the process under watch, OUTPUT is the output received."
       (when (s-present? str)
         (setq result (tabnine-util--read-json str))
 	(unless (equal result 'json-error)
-	  (setq tabnine--response result)
 	  (when (and result (tabnine--valid-response-p result))
-            (setq tabnine--completion-cache-result result)
+	    (when-let*  ((completions (tabnine--filter-completions (plist-get result :results))))
+	      (setq completions (cl-sort completions
+					 (lambda(a b)
+					   (let* ((get-detail-number-fn
+						   (lambda(x)
+						     (when-let* ((md (plist-get x :completion_metadata))
+								 (detail (plist-get md :detail)))
+						       (string-to-number (s-trim detail)))))
+						  (detail-a (funcall get-detail-number-fn a))
+						  (detail-b (funcall get-detail-number-fn b)))
+					     (> detail-a detail-b)))))
+	      (setq tabnine--consecutive-restart-count 0)
+	      (plist-put result :results completions))
+	    (setq tabnine--completion-cache-result result)
 	    (when (and tabnine-mode (equal (point) tabnine--trigger-point))
-	      (tabnine--show-completion-1 result)))))
+	      (tabnine--show-completion-1 result)))
+	  (setq tabnine--response result)))
       (setq ss (cdr ss)))))
 
 ;;
@@ -800,8 +815,7 @@ PROCESS is the process under watch, OUTPUT is the output received."
     (let* ((cache tabnine--completion-cache)
 	   (correlation_id (plist-get result :correlation_id))
 	   (old_prefix (plist-get cache :old_prefix))
-	   (completions (plist-get cache :results))
-	   (completions (tabnine--filter-completions completions)))
+	   (completions (plist-get cache :results)))
       (cond ((seq-empty-p completions)
 	     (message "No completion is available."))
 	    ((= (length completions) 1)
@@ -923,7 +937,7 @@ Use TRANSFORM-FN to transform completion if provided."
 	(delete-region (point)
 		       (min (+ (point) (length old_suffix))
 			    (point-max))))
-      (if (eq major-mode 'vterm-mode)
+      (if (derived-mode-p 'vterm-mode)
 	  (progn
 	    (vterm-delete-region start end)
 	    (vterm-insert t-completion)))
@@ -962,7 +976,6 @@ Use TRANSFORM-FN to transform completion if provided."
   "Show completion result after first get RESPONSE."
   (when-let* ((result response)
 	      (completions (plist-get result :results))
-	      (completions (tabnine--filter-completions completions))
 	      (completion (if (seq-empty-p completions) nil (seq-elt completions 0)))
 	      (old_prefix (plist-get result :old_prefix))
 	      (new_prefix (plist-get completion :new_prefix))
@@ -1244,18 +1257,11 @@ command that triggered `post-command-hook'."
 (defun tabnine--candidates (response)
   "Get candidates for RESPONSE."
   (when (tabnine--response-display-with-capf-p response)
-    (let ((candidates (tabnine--construct-candidates
-		       (plist-get response :results)
-		       #'tabnine--construct-candidate-generic)))
-      (cl-sort candidates
-	       (lambda(a b)
-		 (let* ((get-candidate-detail-number-fn
-			 (lambda(x)
-			   (let ((detail (get-text-property 0 'detail x)))
-			     (string-to-number (s-trim detail)))))
-			(detail-a (funcall get-candidate-detail-number-fn a))
-			(detail-b (funcall get-candidate-detail-number-fn b)))
-		   (> detail-a detail-b)))))))
+    (let* ((result response)
+	   (completions (plist-get result :results)))
+      (tabnine--construct-candidates
+			completions
+			#'tabnine--construct-candidate-generic))))
 
 (defun tabnine--post-completion (candidate)
   "Replace old suffix with new suffix for CANDIDATE."
