@@ -1,5 +1,5 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "tabnine" "0.1.0"
+(define-package "tabnine" "20250102.1608"
   "An unofficial TabNine package with TabNine Chat supported."
   '((emacs        "27.1")
     (dash         "2.16.0")
@@ -8,8 +8,8 @@
     (language-id  "0.5.1")
     (transient    "0.4.0"))
   :url "https://github.com/shuxiao9058/tabnine/"
-  :commit "64b1556e5acd165112827e4219d29cb0f5d24c7b"
-  :revdesc "64b1556e5acd"
+  :commit "7c103aa9e1dd46e8507d341fee42ce30417e69f5"
+  :revdesc "7c103aa9e1dd"
   :keywords '("convenience")
   :authors '(("Aaron Ji" . "shuxiao9058@gmail.com")
              ("Tommy Xiang" . "tommyx058@gmail.com")
